﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;
using System.IO;
using System.Diagnostics;

namespace MinecraftInfoLib
{

    /// <summary>
    /// This is an asynchronous class.
    /// 
    /// TODO: Remove all Console.WriteLine messages and replace them with something useful.
    /// </summary>
    public class ChatClient
    {

        #region Events

        public event EventHandler<LoggedInEventArgs> LoggedIn;
        public event EventHandler<ChatMessageEventArgs> ChatMessage;

        #endregion

        #region Constants

        private const int protocolVersion = 29; // 1.2

        private const byte packet_KeepAlive = 0x00;
        private const byte packet_LoginRequest = 0x01;
        private const byte packet_Handshake = 0x02;
        private const byte packet_ChatMessage = 0x03;

        #endregion

        #region Private Fields

        private IPAddress ipAddress;
        private UInt16 port;

        private Socket socket;
        private AuthInfo authInfo;
        private Statuses status;
        private SessionCreator sessionCreator;

        private Encoding stringEncoding;

        #endregion

        #region Properties

        public IPAddress IPAddress
        {
            get { return ipAddress; }
            set
            {
                if (!Connected)
                {
                    ipAddress = value;
                }
            }
        }

        public UInt16 Port
        {
            get { return port; }
            set
            {
                if (!Connected)
                {
                    port = value;
                }
            }
        }

        public AuthInfo AuthInfo
        {
            get { return authInfo; }
            set
            {
                if (!Connected)
                {
                    authInfo = value;
                }
            }
        }

        public bool Connected
        {
            get
            {
                return (socket != null && socket.Connected);
            }
        }

        public Statuses Status
        {
            get
            {
                return status;
            }
        }

        #endregion

        #region Constructors

        public ChatClient()
        {
            stringEncoding = Encoding.GetEncoding("utf-16BE");
            socket = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
            status = Statuses.Asleep;
        }

        public ChatClient(IPAddress ipAddress_, UInt16 port_)
            : this()
        {
            ipAddress = ipAddress_;
            port = port_;
        }

        public ChatClient(IPAddress ipAddress_)
            : this(ipAddress_, 25565)
        {
        }

        public ChatClient(string ipAddress_)
            : this()
        {
            string[] splitted = ipAddress_.Split(new char[] { ':' }, StringSplitOptions.RemoveEmptyEntries);
            ipAddress = IPAddress.Parse(splitted[0]);

            if (splitted.Length > 1)
            {
                port = UInt16.Parse(splitted[1]);
            }
            else
            {
                port = 25565;
            }
        }

        #endregion

        #region Instance Methods

        public void Connect()
        {
            if (authInfo.ResponseType != AuthInfo.ResponseTypes.Success)
            {
                throw new Exception("AuthInfo: Invalid data");
            }
            if (Connected)
            {
                throw new InvalidOperationException("Already connected.");
            }

            status = Statuses.Connecting;
            socket.BeginConnect(new IPEndPoint(ipAddress, port), socket_ConnectCallback, null);
        }

        public void Disconnect()
        {
            if (!Connected)
            {
                throw new InvalidOperationException("Not connected.");
            }

            status = Statuses.Asleep;
            socket.Disconnect(false);
            socket.Dispose();
        }

        protected virtual void OnLoggedIn(LoggedInEventArgs e)
        {
            EventHandler<LoggedInEventArgs> copy = LoggedIn;
            if (copy != null)
            {
                copy.Invoke(this, e);
            }
        }

        protected virtual void OnChatMessage(ChatMessageEventArgs e)
        {
            EventHandler<ChatMessageEventArgs> copy = ChatMessage;
            if (copy != null)
            {
                copy.Invoke(this, e);
            }
        }

        private void socket_ConnectCallback(IAsyncResult result_)
        {
            socket.EndConnect(result_);

            string handshakeString = String.Format(
                "{0};{1}:{2}",
                authInfo.Username, ipAddress, port);

            byte[] data;

            using (Packet packet = new Packet())
            {
                packet.Write(packet_Handshake);
                packet.Write(handshakeString);

                data = packet.ToArray();
            }

            status = Statuses.Handshaking;
            socket.BeginSend(data, 0, data.Length, SocketFlags.None, socket_SendCallback, null);

            StartReceiving();
        }

        private void StartReceiving()
        {
            byte[] buffer = new byte[1024];
            socket.BeginReceive(buffer, 0, buffer.Length, SocketFlags.None, socket_ReceiveCallback, buffer);
        }

        private void socket_SendCallback(IAsyncResult result_)
        {
            // maybe something will go here ... or not
        }

        public void SendChatMessage(string message_)
        {
            Packet packet = new Packet();
            packet.Write(packet_ChatMessage);
            packet.Write(message_);

            byte[] data = packet.ToArray();
            packet.Dispose();

            socket.BeginSend(data, 0, data.Length, SocketFlags.None, socket_SendCallback, null);
        }

        private void socket_ReceiveCallback(IAsyncResult result_)
        {
            byte[] buffer = (byte[])result_.AsyncState;
            int length = socket.EndReceive(result_);

            if (status == Statuses.Handshaking)
            {
                byte[] hashBytes = new byte[length - 3];
                Array.Copy(buffer, 3, hashBytes, 0, length - 3);
                string hash = stringEncoding.GetString(hashBytes);

                status = Statuses.CreatingSession;
                sessionCreator = new SessionCreator(authInfo, hash);
                sessionCreator.BeginLogin(sessionCreator_LoginCallback);
            }
            else if (status == Statuses.LoggingIn)
            {
                Console.WriteLine("Tight bro, we made it! Or ... at least I think we did ...");

                using (Packet packet = new Packet(buffer))
                {
                    if (packet.ReadByte() == packet_LoginRequest)
                    {
                        Console.WriteLine("We actually made it! Oh mah gaaawd ...");

                        int entityId = packet.ReadInt();
                        packet.ReadString(); // get rid of unused string
                        string levelType = packet.ReadString();
                        int serverMode = packet.ReadInt();
                        int dimension = packet.ReadInt();
                        byte difficulty = packet.ReadByte();
                        packet.ReadByte(); // get rid of unused byte
                        byte maxPlayers = packet.ReadByte();

                        ServerLoginInfo loginInfo = new ServerLoginInfo(
                                entityId, levelType, serverMode, dimension, difficulty, maxPlayers
                                );                     

                        status = Statuses.LoggedIn;

                        OnLoggedIn(new LoggedInEventArgs(loginInfo));
                    }
                    else
                    {
                        Console.WriteLine("Sorry bro, false alarm. :( Keep trying tho!");
                    }
                }
            }
            else if (status == Statuses.LoggedIn)
            {
                using (Packet packet = new Packet(buffer))
                {
                    byte packetId = packet.ReadByte();

                    if (packetId == packet_KeepAlive)
                    {
                        int id = packet.ReadInt();

                        using (Packet newPacket = new Packet())
                        {
                            newPacket.Write(packet_KeepAlive);
                            newPacket.Write(id);

                            byte[] data = newPacket.ToArray();
                            socket.BeginSend(data, 0, data.Length, SocketFlags.None, socket_SendCallback, null);

                            //Console.WriteLine("Kept ya alive, brotha ... " + id);
                        }
                    }
                    else if (packetId == packet_ChatMessage)
                    {
                        /*string message = packet.ReadString();
                        OnChatMessage(new ChatMessageEventArgs(message));*/

                        Debug.Print("Chat message received (" + length + ")");
                    }

                    Debug.Write(packetId + " ");
                }
            }

            StartReceiving();
        }

        private void sessionCreator_LoginCallback(IAsyncResult result_)
        {
            if (sessionCreator.EndLogin(result_))
            {
                Console.WriteLine("Aight, gettin' serious now.");
                byte[] data;

                using (Packet packet = new Packet())
                {
                    packet.Write(packet_LoginRequest);
                    packet.Write(protocolVersion);
                    packet.Write(authInfo.Username);

                    // these are unused
                    packet.Write("");
                    packet.Write((int)0);
                    packet.Write((int)0);
                    packet.Write((byte)0);
                    packet.Write((byte)0);
                    packet.Write((byte)0);

                    data = packet.ToArray();
                }

                status = Statuses.LoggingIn;
                socket.BeginSend(data, 0, data.Length, SocketFlags.None, socket_SendCallback, null);
            }
        }

        #endregion

        #region Nested types

        public enum Statuses
        {
            Asleep,
            Connecting,
            Handshaking,
            CreatingSession,
            LoggingIn,
            LoggedIn
        }

        #endregion

    }
}
