﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.Net.Sockets;

namespace MinecraftInfoLib
{
    public class ServerInfoRetriever : IDisposable
    {

        #region Private Fields

        private IPAddress ipAddress;
        private UInt16 port;

        private const byte packet_ServerListPing = 0xFE;
        private const byte packet_Kick = 0xFF;

        private Func<ServerInfo> fetchDelegate;

        private object accessLock = new object();

        #endregion

        #region Properties

        public IPAddress IPAddress
        {
            get { return ipAddress; }
            set
            {
                lock (accessLock)
                {
                    ipAddress = value;
                }
            }
        }

        public UInt16 Port
        {
            get { return port; }
            set 
            {
                lock (accessLock)
                {
                    port = value;
                }
            }
        }

        #endregion

        #region Constructors

        public ServerInfoRetriever()
        {
            fetchDelegate = Fetch;
        }

        public ServerInfoRetriever(IPAddress ipAddress_, UInt16 port_)
            : this()
        {
            ipAddress = ipAddress_;
            port = port_;
        }

        public ServerInfoRetriever(IPAddress ipAddress_)
            : this(ipAddress_, 25565)
        { }

        public ServerInfoRetriever(string ipString_, UInt16 port_)
            : this(IPAddress.Parse(ipString_), port_)
        { }

        public ServerInfoRetriever(string ipString_)
            : this()
        {
            string[] parts = ipString_.Split(':');

            if (parts.Length > 2)
            {
                throw new ArgumentException("ipString_: Invalid format.");
            }

            ipAddress = IPAddress.Parse(parts[0]);
            if (parts.Length == 2)
            {
                port = UInt16.Parse(parts[1]);
            }
            else
            {
                port = 25565;
            }
        }

        #endregion

        #region Instance Methods

        /// <summary>
        /// Tries to fetch information.
        /// </summary>
        /// <returns>"ServerInfo" object containing information about the server. "null" if unsuccessful.</returns>
        public ServerInfo Fetch()
        {
            byte[] buffer = new byte[1024];
            int length = 0;

            using (Socket socket = new Socket(
                AddressFamily.InterNetwork,
                SocketType.Stream,
                ProtocolType.Tcp))
            {
                lock (accessLock)
                {
                    socket.Connect(ipAddress, port);
                }
                socket.Send(new byte[] { packet_ServerListPing });
                length = socket.Receive(buffer);
                socket.Disconnect(false);
            }

            if (buffer[0] != packet_Kick)
            {
                return null;
            }
            else
            {
                Encoding enc = Encoding.GetEncoding("utf-16BE");

                byte[] information = new byte[length - 3];
                Array.Copy(buffer, 3, information, 0, length - 3);

                Console.WriteLine(BitConverter.ToString(information));

                string informationString = Encoding.GetEncoding("utf-16BE").GetString(information);

                string[] informationSplitted = informationString.Split('§');

                if (informationSplitted.Length != 3)
                {
                    return null;
                }

                return new ServerInfo(
                    Int32.Parse(informationSplitted[1]), 
                    Int32.Parse(informationSplitted[2]), 
                    informationSplitted[0]); 
            }
        }

        public IAsyncResult BeginFetch(AsyncCallback callback_, object state_)
        {
            return fetchDelegate.BeginInvoke(callback_, state_);
        }

        public ServerInfo EndFetch(IAsyncResult result_)
        {
            return fetchDelegate.EndInvoke(result_);
        }

        public void Dispose()
        {
            // not necessary yet
        }

        #endregion

    }

    public class ServerInfo
    {

        #region Private Fields

        private readonly int playerCount;
        private readonly int maxPlayerCount;

        private readonly string description;

        #endregion

        #region Properties

        public int PlayerCount
        {
            get { return playerCount; }
        }
        
        public int MaxPlayerCount
        {
            get { return maxPlayerCount; }
        }

        public string Description 
        {
            get { return description; }
        }

        #endregion

        #region Constructors

        internal ServerInfo(int playerCount_, int maxPlayerCount_, string description_)
        {
            playerCount = playerCount_;
            maxPlayerCount = maxPlayerCount_;
            description = description_;
        }

        #endregion

    }
}
